import Flutter
import AVFoundation

extension Data {

    init<T>(fromArray values: [T]) {
        self = values.withUnsafeBytes { Data($0) }
    }

    func toArray<T>(type: T.Type) -> [T] where T: ExpressibleByIntegerLiteral {
        var array = Array<T>(repeating: 0, count: self.count/MemoryLayout<T>.stride)
        _ = array.withUnsafeMutableBytes { copyBytes(to: $0) }
        return array
    }
}

class _Constants {
    static let workspaceSamples = 100000;
    static let defaultFrameDuration =  0.003;
    static let defaultMaxFrameJitter = 4.0;
    static let processingQueueName = "SwiftAudioIoPluginQueue"
}

enum Methods: String {
    case start
    case stop
    case requestFrameDuration
    case getFrameDuration
    case requestFormat
    case getFormat
}

enum Channels: String {
    case inputChannelName = "com.wearemobilefirst.audio_io.inputAudio";
    case outputChannelName = "com.wearemobilefirst.audio_io.outputAudio";
    case methodChannelName = "com.wearemobilefirst.audio_io";
}

enum AudioDataTypes:String {
    case double
    case float
    case int
}

class _AudioFormat {
    static let sampleRate = "sampleRate"
    static let dataType =  "type"
    static let channels = "channels"
    static let input =  "input"
    static let output = "output"
}
let standardFormat = AVAudioFormat(commonFormat: .pcmFormatFloat64, sampleRate: 44100, channels: 1, interleaved:false)

public class SwiftAudioIoPlugin: NSObject, FlutterPlugin {
    

    let engine = AVAudioEngine()
    var outputConverter = AVAudioMixerNode()
    var _binaryMessenger: FlutterBinaryMessenger?
    var _frameDuration =  _Constants.defaultFrameDuration
    var buffer = RingBuffer<Double>(count:0)
    let maxFrameJitter = _Constants.defaultMaxFrameJitter
    let queue = DispatchQueue(label: _Constants.processingQueueName)
    
    private lazy var sourceNode = AVAudioSourceNode(format: standardFormat! , renderBlock: { (_, _, frameCount, audioBufferList) -> OSStatus in
        let ablPointer = UnsafeMutableAudioBufferListPointer(audioBufferList)
        self.queue.sync() {
            for frame in 0..<Int(frameCount) {
                for buffer in ablPointer {
                    let buf: UnsafeMutableBufferPointer<Double> = UnsafeMutableBufferPointer(buffer)
                    buf[frame] = self.buffer.read() ?? 0
                }
            }
        }
        return noErr
        })
    
    private lazy var sinkNode = AVAudioSinkNode() { (timeStamp, frames, audioBufferList) ->
        OSStatus in
        let sampleCount = Int(frames)
        var doubleSamples = [Double](repeating: 0.0, count: sampleCount)
               let ptr = audioBufferList.pointee.mBuffers.mData?.assumingMemoryBound(to: Float.self)
               let unsafePtr = UnsafeBufferPointer(start: ptr, count: sampleCount);
               var frame = 0;
               for element in unsafePtr {
                   doubleSamples[frame] = Double(element)
                   frame+=1
               }
        let data = Data.init(fromArray:doubleSamples)
        self._binaryMessenger?.send(onChannel: Channels.inputChannelName.rawValue, message: data)
               return noErr
    }
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: Channels.methodChannelName.rawValue, binaryMessenger: registrar.messenger())
        let instance = SwiftAudioIoPlugin()
        registrar.addMethodCallDelegate(instance, channel: channel)
        instance._binaryMessenger = registrar.messenger()
        instance._binaryMessenger?.setMessageHandlerOnChannel(Channels.outputChannelName.rawValue, binaryMessageHandler: { (data, _ ) in
            guard let data = data else{
                return
            };
            let doubles: [Double] = data.toArray(type: Double.self)
            instance.queue.sync() {
                _ = instance.buffer.writeBlock(doubles)
            }
        })
        
        NotificationCenter.default.addObserver(instance, selector: #selector(handleRouteChange), name: AVAudioSession.routeChangeNotification, object: nil)
        NotificationCenter.default.addObserver(instance, selector: #selector(handleInterruption), name: AVAudioSession.interruptionNotification, object: nil)
      
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch call.method {
        case Methods.start.rawValue :
            self.start()
            result(nil)
            break;
        case Methods.stop.rawValue:
            self.stop()
            result(nil)
            break;
        case Methods.requestFrameDuration.rawValue:
            if let requested = call.arguments as? Double {
                _frameDuration = requested
            }
            result(nil)
            break;
        case Methods.getFrameDuration.rawValue:
            result(_frameDuration)
            break;
        case Methods.getFormat.rawValue:
            result(getFormat())
            break;
        default:
            result(FlutterMethodNotImplemented)
        }
    }
    
    public func stop() {
        engine.pause()
    }
    
    public func start() {
        buffer = RingBuffer<Double>(count: Int(_frameDuration * outputSampleRate() * maxFrameJitter));
        // Setup AVAudioSession
        do {
            try AVAudioSession.sharedInstance().setCategory(AVAudioSession.Category.playAndRecord, options: [AVAudioSession.CategoryOptions.allowBluetoothA2DP])
            try AVAudioSession.sharedInstance().setPreferredIOBufferDuration(_frameDuration)
        } catch {
            assertionFailure("AVAudioSession setup error: \(error)")
        }
        outputConverter.outputVolume = 1.0
        
        // Connect nodes
        attach()
        
        // Start engine
        do {
            try engine.start()
        } catch {
            assertionFailure("AVAudioEngine start error: \(error)")
        }
    }
    
    @objc func handleInterruption(notification: NSNotification) {
        guard let userInfo = notification.userInfo,
            let typeValue = userInfo[AVAudioSessionInterruptionTypeKey] as? UInt,
            let type = AVAudioSession.InterruptionType(rawValue: typeValue) else {
                return
        }
        switch type {
        case .began:
            engine.pause()
        case .ended:
            do {
                try engine.start()
            } catch {
                assertionFailure("AVAudioEngine start error: \(error)")
            }
        default: ()
        }
    }
    
    @objc func handleRouteChange(notification: NSNotification) {
        guard let userInfo = notification.userInfo,
            let reasonValue = userInfo[AVAudioSessionRouteChangeReasonKey] as? UInt,
            let reason = AVAudioSession.RouteChangeReason(rawValue: reasonValue) else {
                return
        }
        switch reason {
        case .newDeviceAvailable, .oldDeviceUnavailable, .wakeFromSleep:
            self.resetAudio();
        default: ()
        }
    }
    
    public func resetAudio() {
        engine.pause()
        engine.reset()
        engine.detach(outputConverter)
        engine.detach(sinkNode)
        engine.detach(sourceNode)
        engine.disconnectNodeInput(engine.inputNode)
        buffer.clear()
        start();
    }
    
    public func inputSampleRate() -> Double {
        let input = engine.inputNode
        let inputFormat = input.outputFormat(forBus: 0)
        return inputFormat.sampleRate
    }
    
    public func outputSampleRate() -> Double {
        let output = engine.mainMixerNode
        let outputFormat = output.outputFormat(forBus: 0)
        return outputFormat.sampleRate
    }
    
    public func attach() {
        // Setup engine and node instances
        let input = engine.inputNode
        let output = engine.mainMixerNode
        let outputFormat = output.inputFormat(forBus: 0)
        
        // Connect nodes
        engine.attach(outputConverter)
        engine.attach(sinkNode)
        engine.attach(sourceNode)
        engine.connect(input, to: sinkNode, format: nil)
        engine.connect(sourceNode, to: outputConverter, format: nil)
        engine.connect(outputConverter, to: output, format: outputFormat)
    }
    
    public func getFormat() -> [String:Any] {
        let input = engine.inputNode
        let output = engine.mainMixerNode
        let inputFormat = input.outputFormat(forBus: 0)
        let outputFormat = output.inputFormat(forBus: 0)
        
        let inputDesc: [String:Any] = [_AudioFormat.dataType: AudioDataTypes.double.rawValue,
                                       _AudioFormat.channels: inputFormat.channelCount,
                                       _AudioFormat.sampleRate: inputFormat.sampleRate]
        
        let outputDesc: [String:Any] = [_AudioFormat.dataType: AudioDataTypes.double.rawValue,
                                        _AudioFormat.channels: outputFormat.channelCount,
                                        _AudioFormat.sampleRate: outputFormat.sampleRate]
        
        return [_AudioFormat.input: inputDesc, _AudioFormat.output: outputDesc]
    }
    
    
}



public struct RingBuffer<T> {
    fileprivate var array: [T?]
    fileprivate var readIndex = 0
    fileprivate var writeIndex = 0
    
    public init(count: Int) {
        array = [T?](repeating: nil, count: count)
    }
    
    public mutating func write(_ element: T) -> Bool {
        if !isFull {
            array[writeIndex % array.count] = element
            writeIndex += 1
            return true
        } else {
            return false
        }
    }
    
    public mutating func writeBlock(_ block: [T?]) -> Bool {
        let count = block.count
        if(availableSpaceForWriting >= count){
            for i in 0..<count {
                array[(writeIndex+i) % array.count] =  block[i];
            }
            writeIndex += count
            return true
        }
        return false
    }
    
    public mutating func read() -> T? {
        if !isEmpty {
            let element = array[readIndex % array.count]
            readIndex += 1
            return element
        } else {
            return nil
        }
    }
    
    public mutating func readBlock(count: Int) -> [T?]? {
        if(availableSpaceForReading >= count){
            var result = [T?](repeating: nil, count: count)
            for i in 0..<count {
                result[i] =  array[(readIndex+i) % array.count];
            }
            readIndex += count
            return result
        }
        return nil
    }
    
    public mutating func clear() {
        readIndex = 0;
        writeIndex = 0;
    }
    
    fileprivate var availableSpaceForReading: Int {
        return writeIndex - readIndex
    }
    
    public var isEmpty: Bool {
        return availableSpaceForReading == 0
    }
    
    fileprivate var availableSpaceForWriting: Int {
        return array.count - availableSpaceForReading
    }
    
    public var isFull: Bool {
        return availableSpaceForWriting == 0
    }
}


